close all
clear all
clc
tic

%% INPUT
    % Time series
        T = 2000;                    % Simulation time
        dt = .05;                    % Time step
        Ns = T/dt+1;                  % Number of samplings
        t = (0:dt:T);                 % Time vector
        n =T / 2;                      % Number of wave components
        phase = 2*pi*rand(1,n);       % Random phase (for time series of wave spectrum)
   
    % Damage Calculation
        beta = (2:0.05:5);
        %beta = 5;
        nb = length(beta);
        
%% Generate time series from spectrum

    %  Frequency vector
        f1 = 0;
        f2 = 1/2/dt;
        df =(f2-f1)/(n-1);
        f = (f1:df:f2) + df.*rand(1,n);
        
    %   system
        b = f / (2.1 / 2 / pi);
        zeta = 0.1;
        k = 1.8e6;
        S0 = (8e5)^2*0.05/pi;
        AR1 = 0.;
        sigma_w = 5e8;
    % White Noise Response Spectrum
        S = S0 * 2 * pi * (k.^2*((1-b.^2).^2+(2*zeta*b).^2)).^(-1);
        S_AR =  2 * sigma_w^2 * 0.05 ./ (1 - 2 * AR1 * cos(f.*2*pi * 0.05) + AR1^2) ./ (k.^2*((1-b.^2).^2+(2*zeta*b).^2));
        A = sqrt(2*S_AR*df);   % Amplitude for each wave component
    % Wave elevation time series
        for i = 1:length(t)
            x(i) = sum(A .* cos(2 * pi * f*t(i) + phase));
        end
        disp('Signal Generated, log size:')
        disp(log10(size(t,2)))
      
%% Damage
        % t = gather(t);
        % x = gather(e);
        
    % RFC
        tp = dat2tp([t',x']);
        RFC = tp2rfc(tp);
        
    % Total damage - WAFO
        fb = power(ones(1,nb)*2,beta);  % factor to match WAFO cc2dam function
        DRFC = cc2dam(RFC,beta);
        dRFC = fb.*DRFC/T;
        
    % Total damage - Script
        nRFC = length(RFC);
        D = sum(power(repmat((RFC(:,2)-RFC(:,1)),1,nb), repmat(beta,nRFC,1)))/T;
        
    % Save to csv
        csvwrite('Damage.csv', D)
        disp('Damage Calculated')

        
%% Rainflow counting distribution
    minRFC = min(RFC(:,1));
    maxRFC = max(RFC(:,2));
    if minRFC < -4 || maxRFC > 4
        disp('u does not fit')
    end
    
    u = (4:-0.1:-4);
    nu = length(u);
    NRFC = zeros(nu);
    
    for ui = 1:nu
        for vi = ui:nu
            for i = 1:nRFC
                if RFC(i,2) > u(ui) && u(vi) > RFC(i,1)
                    NRFC(ui,vi) = NRFC(ui,vi) + 1;
                end
            end
        end
    end
    NRFC = NRFC/T;
    
%% Total Damage - Integration
for bi = 1:nb
    
    clear fdd fd I_1
    
    for ui = 1:nu
        for vi = 1:nu
           fdd(ui,vi) = -beta(bi)*(beta(bi)-1)*(u(ui)-u(vi))^(beta(bi)-2);
           fd(ui,vi) = -beta(bi)*(u(ui)-u(vi))^(beta(bi)-1);
        end
    end
    
    for i = 1:nu
        I_1(i)=trapz(u(i:end),NRFC(i,i:end).*fdd(i,i:end),2);
    end
    
    ED(bi) = -trapz(u,I_1);

end
        
%% Plots

        figure
        plot(beta,D,beta,ED)
        title('Damage intensity as function of \beta')
        legend('PM','Integral')
        xlabel('\beta')
        ylabel('intensity')
        
        % figure
        % ff = linspace(0,1/2/dt,size(pxx,1));
        % semilogy(f, S,ff,pxx, ff, pww)
        % legend('exact', 'welch')
        
toc