  # -*- coding: utf-8 -*-
"""
Created on Sat Apr 1 18:41:58 2017

@author: Odd Eiken
"""
from numba import jit
from cell_discretization import *
from cell_mapping import boundary_matrix
from main import load_sparse_csc
import numpy as np
import plotting as plot
import matplotlib.pyplot as plt
import scipy.linalg as LA
from scipy.ndimage import gaussian_filter 
from scipy.special import gamma

def load_extrema_matrix(filename):
    fn = '../PROGRAM - Eddinburgh/saved/matrix/%s_extrema_full.npz' % (filename)
    loader = np.load(fn)
    mesh = loader['mesh'].item()
    par = loader['par'].item()  # [omega_n, zeta, k, dt, AR1, sigma_w]
    Matrix = loader['Matrix']
    return mesh, par, Matrix


def save_rainflow_intensity_matrix(filename, INT_RFC):
	fn = '../PROGRAM - Eddinburgh/saved/matrix/%s_intrfc.npz' % (filename)
	np.savez(fn, INT_RFC=INT_RFC)
	print 'rainflow intensity matrix saved!'


def load_rainflow_intensity_matrix(filename):
	fn = '../PROGRAM - Eddinburgh/saved/matrix/%s_intrfc.npz' % (filename)
	loader = np.load(fn)
	return loader['INT_RFC']


def power_method(P):
    # compute Pi
    size, size = P.shape
    pi = np.zeros(size)
    pi1 = np.zeros(size)
    pi[0] = 1
    n = LA.norm(pi - pi1, 1)
    i = 0
    while n > 1e-6 and i < 1e5:
        pi1 = np.dot(pi, P)
        pi = np.dot(pi1, P)
        n = LA.norm(pi - pi1, 1)
        i += 1
    return pi


def extrema_transition(file, b_iter=1000):
    '''
    returns extrema transition matrix for x values from X_max -> -X_min
    '''
    mesh, par, P_init = load_sparse_csc(file)
    ndim = len(mesh.keys())
    x = mesh['x']
    v = mesh['v']
    del mesh['x']
    del mesh['v']
    size = [x.size, v.size]
    for key in mesh.keys():
        size.append(mesh[key].size)
    mesh['x'] = x
    mesh['v'] = v
    ncells = np.prod(size)
    n = len(x)
    E = np.zeros((n, n))
    Pb = boundary_matrix(P_init, b_iter, mesh=mesh, par=par)
    v0_idx = np.argmin(abs(v))
    if ndim == 3:
        cx = np.arange(ncells).reshape(size)[:, v0_idx, :].flatten()  # cells along x - axis
    elif ndim == 4:
        cx = np.arange(ncells).reshape(size)[:, v0_idx, :, :].flatten()  # cells along x - axis
    for j in np.arange(n):
        x0_idx = j
        if ndim == 3:
            c0 = np.arange(ncells).reshape(size)[x0_idx, v0_idx, :].flatten()  # cells at starting point
        elif ndim == 4:
            c0 = np.arange(ncells).reshape(size)[x0_idx, v0_idx, :, :].flatten()  # cells at starting point
        p = np.zeros(ncells)
        p[c0] = 1. / len(c0)
        p = P_init * p
        max_it = 1000
        tol = 5.e-5
        PXT = np.zeros((max_it, n))
        if ndim == 3:
        	r_shape = size[0], size[2]
        	axis = [1]
        elif ndim == 4:
        	r_shape = size[0], size[2], size[3]
        	axis = [2, 1]
        for i in xrange(max_it):
            p = Pb * p
            px = p[cx].reshape(r_shape)
            for a in axis:
            	px = np.sum(A, axis=a)
            PXT[i, :] = px
            p[cx] = 0
            if p.sum() < tol:
                PXT = PXT[:i + 1, :]
                break
        PX = np.sum(PXT[np.arange(i), :], axis=0)
        E[:, j] += PX / np.sum(PX)
        print 'iterations: ', i
        print 'cell number: ', j
        print
    fn = '../PROGRAM - Eddinburgh/saved/matrix/%s' % (file)
    np.savez(fn + '_extrema_full', mesh=mesh, par=par, Matrix=E)
    return E


def get_x0_col(p, cx, size):
    ndim = len(size)
    if ndim == 3:
        px = p[cx].reshape(size[0], size[2])
        px = np.sum(px[:, np.arange(size[2])], axis=1)
        return px
    elif ndim == 4:
        px = p[cx].reshape(size[0], size[2], size[3])
        px = np.sum(px[:, :, np.arange(size[3])], axis=2)
        px = np.sum(px[:, np.arange(size[2])], axis=1)
        return px


def find_zero_cells(mesh):
        '''
        returns cells where velocity is zero
        '''
        ndim = len(mesh.keys())
        x = mesh['x']
        v = mesh['v']
        del mesh['x']
        del mesh['v']
        size = [x.size, v.size]
        for key in mesh.keys():
            size.append(mesh[key].size)
        mesh['x'] = x
        mesh['v'] = v
        ncells = np.prod(size)
        v_zero = np.arange(size[1] / 2 - 1 + size[1] % 2, size[1] / 2 + 1)
        C = np.arange(ncells).reshape(size)
        if ndim == 3:
            c_zeros = C[:, v_zero, :].flatten()
        elif ndim == 4:
            c_zeros = C[:, v_zero, :, :].flatten()
        return c_zeros


def plot_cycle(file, x0_idx, b_iter=10):
    '''
    plots a halfs cycle path
    file : transition matrix file name
    x0_idx : index where the half cycle starts
    b_iter : number of iterations in the new boundary matrix
    			default = 10
    '''
    mesh, par, P_init = load_sparse_csc(file)
    ndim = len(mesh.keys())
    x = mesh['x']
    v = mesh['v']
    del mesh['x']
    del mesh['v']
    size = [x.size, v.size]
    for key in mesh.keys():
        size.append(mesh[key].size)
    mesh['x'] = x
    mesh['v'] = v
    dx = x[1] - x[0]
    dv = v[1] - v[0]
    print "x_0 =", x[x0_idx]
    ncells = np.prod(size)
    Pb = boundary_matrix(P_init, b_iter, mesh=mesh, par=par)
    v0_idx = np.argmin(abs(v))
    cx = find_zero_cells(mesh)
    if ndim == 3:
        c0 = np.arange(ncells).reshape(size)[x0_idx, v0_idx, :].flatten()  # cells at starting point
    elif ndim == 4:
        c0 = np.arange(ncells).reshape(size)[x0_idx, v0_idx, :, :].flatten()  # cells at starting point
    p = np.zeros(ncells)
    p[c0] = 1. / len(c0)
    max_it = 100
    tol = 5.e-6  # amount of probability not reached recurent state
    PXT = np.zeros((max_it, len(x)))
    PXV = np.zeros((len(x), len(v))).T
    px = get_x0_col(p, cx, size)
    p1, p2, pxv = plot.data(p, mesh)
    p = P_init * p
    px = get_x0_col(p, cx, size)
    PXT[1, :] = px
    p1, p2, pxv = plot.data(p, mesh)
    PXV += pxv
    p[cx] = 0
    for i in xrange(2, max_it):
        print i - 1
        p = Pb * p
        px = get_x0_col(p, cx, size)
        PXT[i, :] = px
        p1, p2, pxv = plot.data(p, mesh)
        PXV += pxv
        print p.sum()
        p[cx] = 0
        if p.sum() < tol:
            PXT = PXT[:i + 3, :]
            break
    i, j = PXT.shape
    print p.sum()
    PX = np.sum(PXT[np.arange(i), :], axis=0)
    PT = np.sum(PXT[:, np.arange(j)], axis=1)
    print 'px-sum: ', np.sum(PX)
    print 'pt-sum: ', np.sum(PT)
    print 'pxt-sum: ', np.sum(PXT)
    dt = par['dt']
    t = np.arange(i) * dt
    filter = 1
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 5))   # Time and displacement
    PX = gaussian_filter(PX, filter)
    PT = gaussian_filter(PT, filter)
    ax1.plot(x[x0_idx], PX.max() / (x[1] - x[0]) * 0.01, 'bo', label='Start')
    ax1.plot(x, PX / (x[1] - x[0]), label='PDF half-cycle')
    ax1.axis([x[0], x[-1], 0, 1.2 * PX.max() / (x[1] - x[0])])
    ax2.plot(0, PT.max() / dt * 0.01, 'bo')
    ax2.plot(t, PT / dt, label='PDF - halfsycle')
    ax1.set_title('Displacement [m]')
    ax2.set_title('Time [s]')
    ax1.legend(fancybox=True, loc='best')
    ax2.axis([t[0], t[-1], 0, 1.2 * PT.max() / dt])
    plt.tight_layout()
    
    fig, ax = plt.subplots(figsize=(10, 7))  # path
    lim = [x[0], x[-1], v[0], v[-1]]
    xticks = np.linspace(x[0], x[-1], 5)
    yticks = np.linspace(v[0], v[-1], 5)
    ax.vlines(0, v[0], v[-1], color='silver')
    ax.hlines(0, x[0], x[-1], color='silver')
    vmax = PXV.max() / dx / dv * .05
    cax = ax.matshow(PXV / dx / dv, extent=lim, origin='lower', interpolation='bilinear', aspect='auto', cmap='BuPu', vmin=0, vmax=vmax)
    values = (np.arange(5) * vmax) / 4
    cbar = fig.colorbar(cax, ticks=values)
    cbar.ax.set_yticklabels(np.round(values, 1))
    cbar.ax.yaxis.set_ticks_position('right')
    ax.set_xticks(np.round(xticks, 1))
    ax.set_yticks(np.round(yticks, 1))
    ax.set_axisbelow(True)
    ax.set_xlabel('Displacement')
    ax.set_ylabel('Velocity')
    ax.axis([x[0], x[-1], v[0], v[-1]])

    ax.plot(x[x0_idx], 0, 'bo')
    plt.tight_layout()

    fig, ax3 = plt.subplots(figsize=(10, 7))  # time vs displacement
    lim = t[0], t[-1], x[0], x[-1]
    xticks = np.linspace(t[0], t[-1], 5)
    yticks = np.linspace(x[0], x[-1], 5)
    vmax = PXT.max() / dx / dt * .1
    cax = ax3.matshow(PXT.T / dx / dt, extent=lim, origin='lower', interpolation='bilinear', aspect='auto', cmap='BuPu', vmin=0, vmax=vmax)
    values = (np.arange(5) * vmax) / 4
    cbar = fig.colorbar(cax, ticks=values)
    cbar.ax.set_yticklabels(np.round(values, 1))
    cbar.ax.yaxis.set_ticks_position('right')
    ax3.hlines(0, 0, t[-1])
    ax3.plot(0, x[x0_idx], 'bo')
    ax3.set_xlabel('Time [s]')
    ax3.set_ylabel('Displacement')
    ax3.set_xticks(np.round(xticks, 1))
    ax3.set_yticks(np.round(yticks, 1))
    ax3.axis(lim)
    print 'cell number: ', x0_idx
    print 'x = :', x[x0_idx]


@jit
def PT_intensity(E):
	'''
	E : Extrema - to Extrema matrix
	'''
	n, n = E.shape
	p_p = power_method(E)
	p_p = gaussian_filter(p_p, 0)
	P = np.zeros(E.shape)
	INT = np.zeros(E.shape)
	for i in range(n):  # P[u=x_i, v=x_j]
	    P[i, :i] = E[i, :i] * p_p[i]
	for i in range(n):  # p(x>u, y<v)
	    for j in range(i + 1):
	    	INT[i, j] = np.sum(P[i:, :j + 1])
	return INT




def rainflow_int(INT_PT, E):
    '''
    Algorith finding the rainflow intensity matrix
    from the transition matrix and peakt-trough intensity matrix
    '''
    n, n = E.shape
    I = np.zeros((n, n))
    for i in range(n):
        I[i, n - 1 - i] = 1
    I = np.matrix(I)
    M = np.matrix(E)
    E = (I * M * I)     #Ematrix maps with matrix x=[max_x->min_x]
    P = np.zeros(E.shape)
    P_hat = np.zeros(E.shape)
    for i in range(n):
        P[i, i:] = E[i, i:]
        P_hat[i, :i + 1] = E[i, :i + 1]
    INT_PT = I * np.matrix(INT_PT) * I
    INT_RFC = np.zeros(INT_PT.shape)

    @jit
    def A_matrix(i, j, P):
        return np.matrix(P[i:j, i + 1:j + 1])

    @jit
    def B_matrix(i, j, P_hat):
        return np.matrix(P_hat[i + 1:j + 1, i:j])

    @jit
    def p_vector(j, P):
        n, n = P.shape
        p = np.zeros(n)
        l = np.arange(j + 1, n)
        for k in range(n):
            p[k] = np.sum(P[k, l])
        return np.matrix(p).T

    @jit
    def e(i, j, P):
        return p_vector(j, P)[i:j]

    @jit
    def q_vector(i, j, int_pt):
        l = np.arange(i + 1, j + 1)
        return np.matrix(int_pt[i, l - 1] - int_pt[i, l])

    for i in xrange(n):
        print i
        for j in xrange(i, n):
            A = A_matrix(i, j, P)
            B = B_matrix(i, j, P_hat)
            q = q_vector(i, j, INT_PT)
            INT_RFC[i, j] = INT_PT[i, j] + q * B * np.linalg.inv(np.eye(j - i) - A * B) * e(i, j, P)
    return np.array(I * INT_RFC * I)

  
def damage(INT, file):
    INT = np.array(INT)
    mesh, par, E = load_extrema_matrix(file)
    n, n = E.shape
    x = mesh['x']
    dx = x[1] - x[0]

    def fdd(u, v, beta):  #second derivative of damage function
    	return -beta * (beta - 1.) * (u - v)**(beta - 2.)

    steps = 50
    damage = np.zeros(steps)
    beta = np.linspace(2, 5, steps)
    i = 0
    for beta_i in beta:
        FDD = np.zeros((n, n))
        for vi in range(n):
            for ui in range(vi, n):
                u = x[ui]
                v = x[vi]
                FDD[ui, vi] = fdd(u, v, beta_i)
        I = 0  #integral sum
        M = FDD * INT
        for j in range(n):  # summing over all entities in matrix
            for k in range(j):
                I += M[j, k] * dx**2
 
        damage[i] = - I  # Damage for beta_i
        i += 1
    return beta, damage


if __name__ == '__main__':
    
    file = 'WN_fatigue2'
    E = extrema_transition(file, b_iter=100)
    mesh, par, E = load_extrema_matrix(file)
    lim = mesh['x'][0], mesh['x'][-1], mesh['x'][0], mesh['x'][-1]
    E = E.T
    plot_cycle(file, 10, 10)
    fig, ax = plt.subplots(figsize=(10, 7))
    vmax = 0.02
    ndig = 2
    cax = ax.imshow(E, origin='lower', cmap="Greys", vmin=0, vmax=vmax, extent=lim)
    cbar = fig.colorbar(cax, ticks=[0, vmax / 2, vmax])
    c_max = round(vmax, ndig)
    c_mid = c_max / 2.
    cbar.ax.set_yticklabels(['0.0', c_mid, c_max])  # colorbar
    ax.set_xlabel(r"$Peak_{k+1}\quad x_j$")
    ax.set_ylabel(r"$Peak_{k}\quad x_i$")
    a = True
    sigma_x = 1.05299667311
    if a:
        fig, ax = plt.subplots(figsize=(10, 7))
        p = power_method(E)
        p = gaussian_filter(p, 0)
        ax.plot(mesh['x'], p / (mesh['x'][1] - mesh['x'][0]), label='peak distr.')
        xx = np.linspace(0, sigma_x * 4, 1000)
        yy = xx / sigma_x**2 * np.exp(-xx**2 / 2 / sigma_x**2)
        ax.plot(xx, yy / 2, c='r', label='Rayleigh distr.')
        ax.legend(fancybox=True, loc='best')
        ax.set_xlabel('Displacement')
        INT_PT = PT_intensity(E)
        INT_RF = rainflow_int(INT_PT, E); save_rainflow_intensity_matrix(file, INT_RF)
        INT_RF = load_rainflow_intensity_matrix(file)
        fig, ax = plt.subplots(figsize=(10, 7))
        beta, damage = damage(INT_RF, file)
        np.savetxt("damage_markov.csv", damage, delimiter=",")
        ax.plot(beta, damage)
        ax.plot(beta, par["omega_n"]/ 2 / np.pi * (2**(3*beta/2) * (1. / sigma_x**2)**(-beta / 2) * gamma(beta/2. + 1)), "--")
        ax.set_xlabel(r'$\beta$', size=15)
        ax.set_title('Damage Intensity')
        fig, ax = plt.subplots(figsize=(10, 7))
        vmax = .4
        ndig = 2
        lim = mesh['x'][0], mesh['x'][-1], mesh['x'][0], mesh['x'][-1]
        cax = ax.imshow(INT_RF, origin='lower', aspect='auto', interpolation='nearest', cmap='jet', extent=lim, vmax=vmax, vmin=0)
        cbar = fig.colorbar(cax, ticks=[0, vmax / 2, vmax])
        c_max = round(vmax, ndig)
        c_mid = c_max / 2.
        cbar.ax.set_yticklabels(['0.0', c_mid, c_max])  # vertically oriented colorbar
        ax.set_xlabel("v - levels")
        ax.set_ylabel("u - levels")
    plt.show()
