# -*- coding: utf-8 -*-
"""
Created on Thu May 11 11:59:04 2017

@author: Odd Eiken
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import LinearLocator, FormatStrFormatter
from mpl_toolkits.mplot3d import Axes3D
import scipy.ndimage
from main import load_sparse_csc


def map_to_same(file):
    '''
    Takes a Transition matrix and finds how much probability is mapped into the same (x, v) - state
    '''
    mesh, par, P = load_sparse_csc(file)
    x = mesh['x']
    v = mesh['v']
    F = mesh['F']
    indptr = P.indptr
    indices = P.indices
    data = P.data
    p = np.zeros(x.size * v.size)
    n_F = F.size
    for i in xrange(len(indptr) - 1):
        L = indptr[i]
        U = indptr[i + 1]
        same_state = np.arange(n_F) + i - i % n_F
        ind = indices[L: U]
        dat = data[L: U]
        for j in xrange(U - L):
            if ind[j] in same_state:
                p[i / n_F] += dat[j] / n_F
    return p.reshape(x.size, v.size)


if __name__ == '__main__':
    file = 'WN'
    PXV = map_to_same(file)
    PXV = PXV.T
    mesh, par, P = load_sparse_csc(file)
    x = mesh['x']
    v = mesh['v']
    F = mesh['F']
    PX = np.sum(PXV[:, np.arange(v.size)], axis=1) / v.size
    PV = np.sum(PXV[np.arange(x.size), :], axis=0) / x.size
    fig, (ax1, ax2) = plt.subplots(2)
    ax1.plot(PX)
    ax2.plot(PV)
    fig, ax = plt.subplots()
    lim = [round(x[0], 2), round(x[-1], 2), round(v[0], 2), round(v[-1], 2)]
    print lim
    cax = ax.matshow(PXV, aspect='auto', interpolation='nearest', cmap='jet', extent=lim, vmin=0, vmax=1)
    cbar = fig.colorbar(cax, ticks=[0, 0.5, 1])
    cbar.ax.set_yticklabels(['0.0', '0.5', '1.0'])  # vertically oriented colorbar
    # 3D plot
    X, V = np.meshgrid(x, v)
    Z = PXV
    rho = 0
    Z = scipy.ndimage.gaussian_filter(Z, rho)

    region = np.s_[5:50, 5:50]
    fig, ax = plt.subplots(subplot_kw=dict(projection='3d'), figsize=(10, 10))
    surf = ax.plot_surface(V, X, Z, rstride=1, cstride=1, cmap='jet', alpha=1,
                           linewidth=0.0, antialiased=False, shade=True)
    surf.set_clim(vmin=0, vmax=Z.max())
    ax.set_zlim(0, Z.max())
    ax.zaxis.set_major_locator(LinearLocator(5))
    ax.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))
    ax.set_title('Joint probability Density Plot')
    ax.set_xlabel('Velocity')
    ax.set_ylabel('Displacement')
    ax.set_zlabel('probability density')
    plt.show()
