# -*- coding: utf-8 -*-
"""
Created on Thu May 11 09:57:37 2017

@author: Odd Eiken
"""
import numpy as np
import scipy.stats as ss
from scipy.integrate import quad
import matplotlib.pyplot as plt
import cell_discretization as cell
from scipy.signal import welch
from main import load_array
import plotting as plot


def FRF(omega_f):
    omega_n, zeta, k = cell.system()
    beta = omega_f / omega_n
    return 1. / (k * ((1. - beta**2)**2 + (2 * zeta * beta)**2)**.5)


def response_displ_AR1(omega, dt, AR1, sigma_w):
    omega_n, zeta, k = cell.system()
    beta = omega / omega_n
    return sigma_w**2 * dt / np.pi / (1. - 2. * AR1 * np.cos(omega * dt) + AR1**2) / (k**2 * ((1. - beta**2)**2 + (2 * zeta * beta)**2))


def response_vel_AR1(omega, dt, AR1, sigma_w):
    omega_n, zeta, k = cell.system()
    beta = omega / omega_n
    return omega**2 * sigma_w**2 * dt / np.pi / (1. - 2. * AR1 * np.cos(omega * dt) + AR1**2) / (k**2 * ((1. - beta**2)**2 + (2 * zeta * beta)**2))


def response_displ_AR2(omega, dt, AR1, AR2, sigma_w):
    omega_n, zeta, k = cell.system()
    beta = omega / omega_n
    return sigma_w**2 * dt / np.pi / (1. + AR1**2 + AR2**2 - 2. * AR1 * (1. - AR2) * np.cos(omega * dt) - 2 * AR2 * np.cos(2 * omega * dt)) / (k**2 * ((1. - beta**2)**2 + (2 * zeta * beta)**2))


def response_vel_AR2(omega, dt, AR1, AR2, sigma_w):
    omega_n, zeta, k = cell.system()
    beta = omega / omega_n
    return omega**2 * sigma_w**2 * dt / np.pi / (1. + AR1**2 + AR2**2 - 2. * AR1 * (1. - AR2) * np.cos(omega * dt) - 2 * AR2 * np.cos(2 * omega * dt)) / (k**2 * ((1. - beta**2)**2 + (2 * zeta * beta)**2))


def RKN(par, x, v, F0, F1, F2):
    omega_n, zeta, k, h = par
    K = 0.5 * h * omega_n**2
    k1 = K * (F0 / k - 2 * zeta / omega_n * v - x)
    k2 = K * (F1 / k - 2 * zeta / omega_n * (v + k1) - (x + 0.5 * h * (v + .5 * k1)))
    k3 = K * (F1 / k - 2 * zeta / omega_n * (v + k2) - (x + 0.5 * h * (v + .5 * k1)))
    k4 = K * (F2 / k - 2 * zeta / omega_n * (v + 2 * k3) - (x + h * (v + k3)))
    xn = x + h * (v + (k1 + k2 + k3) / 3.)
    vn = v + (k1 + 2 * k2 + 2 * k3 + k4) / 3.
    return xn, vn


def simulate(N_it, noise=1):
	'''
	noise = 1 : pick values from gaussian distribution
	noise = 2 : pick values from discrete triplet distribution
	'''
	omega_n, zeta, k = cell.system()
	par = cell.par()
	dt = par['dt']
	sigma_x = 0
	sigma_x_meas = 0
	sigma_v_meas = 0
	sigma_v = 0
	sigma_F = 0
	print par
	H = np.zeros((200, 200))
	pX = np.zeros(200)
	pV = np.zeros(200)
	pF = np.zeros(100)

	loops = 1
	for i in range(loops):
		X = np.zeros(N_it)
		V = np.zeros(N_it)
		print i

		try:
			AR1 = par['AR'][0]
			sigma_w = par['sigma_w']
			FAR1 = np.zeros(N_it * 2)
			if len(par['AR']) == 2:
				AR2 = par['AR'][1]
				FAR2 = np.zeros(N_it * 2)
				FAR1 = 0
				var_displ, err = quad(response_displ_AR2, 0, np.pi / dt, args=(dt, AR1, AR2, sigma_w))
				var_vel, err = quad(response_vel_AR2, 0, np.pi / dt, args=(dt, AR1, AR2, sigma_w))
				sigma_x = var_displ**.5
				sigma_v = var_vel**.5
				sigma_F = ((1. - AR2) / (1. + AR2) * sigma_w**2 / ((1. - AR2)**2 - AR1**2))**.5
				print "sigma_F: ", sigma_F
				if noise == 1:
					for i in xrange(1, N_it):
						FAR2[i * 2] = FAR2[2 * i - 2] * AR1 + FAR2[2 * i - 4] * AR2 + np.random.normal() * sigma_w
						FAR2[i * 2 - 1] = 0.5 * (FAR2[2 * i] + FAR2[2 * i - 2])
				if noise == 2:
					W, pW = cell.AR_NOISE()
					for i in xrange(1, N_it):
						FAR2[i * 2] = FAR2[2 * i - 2] * AR1 + FAR2[2 * i - 4] * AR2 + np.random.choice(W, p=pW)
						FAR2[i * 2 - 1] = 0.5 * (FAR2[2 * i] + FAR2[2 * i - 2])
			else:
				AR2 = 0
				FAR2 = 0
				var_displ, err = quad(response_displ_AR1, 0, np.pi / dt, args=(dt, AR1, sigma_w))
				var_vel, err = quad(response_vel_AR1, 0, np.pi / dt, args=(dt, AR1, sigma_w))
				sigma_x = var_displ**.5
				sigma_v = var_vel**.5
				sigma_F = sigma_w / (1 - AR1**2)**.5
				if noise == 1:
					for i in xrange(1, N_it):
						FAR1[i * 2] = FAR1[2 * i - 2] * AR1 + np.random.normal() * sigma_w
						FAR1[i * 2 - 1] = 0.5 * (FAR1[2 * i] + FAR1[2 * i - 2])
				if noise == 2:
					W, pW = cell.AR_NOISE()
					for i in xrange(1, N_it):
						FAR1[i * 2] = FAR1[2 * i - 2] * AR1 + np.random.choice(W, p=pW)
						FAR1[i * 2 - 1] = 0.5 * (FAR1[2 * i] + FAR1[2 * i - 2])
			
		except KeyError:
			AR1 = 0
			sigma_w = 0
			AR2 = 0
			FAR1 = 0
			FAR2 = 0

		
		if 'F0' in par:

			F0 = par['F0'][0]
			omega_f0 = par['omega_f'][0]
			F0 = F0 * np.sin(np.arange(N_it * 2) * dt / 2 * omega_f0 + np.random.rand() * 2 * np.pi)
			sigma_x += FRF(omega_f0) / 2 * par['F0'][0]
			sigma_v += FRF(omega_f0) * omega_f0 / 2 * par['F0'][0]
			sigma_F += par['F0'][0] / 2

			if len(par['F0']) == 2:
				F1 = par['F0'][1]
				omega_f1 = par['omega_f'][1]
				F1 = F1 * np.sin(np.arange(N_it * 2) * dt / 2 * omega_f1 + np.random.rand() * 2 * np.pi)
				sigma_x += FRF(omega_f1) / 2 * par['F0'][0]
				sigma_v += FRF(omega_f1) * omega_f0 / 2 * par['F0'][0]
				sigma_F += par['F0'][1] / 2
			else:
				F1 = 0
		else:
			F0 = 0
			F1 = 0
			omega_f1 = 0
		F = FAR1 + FAR2 + F0 + F1
		parRKN = omega_n, zeta, k, dt
		for i in xrange(1, N_it):
			X[i], V[i] = RKN(parRKN, X[i - 1], V[i - 1], F[2 * (i - 1)], F[2 * (i - 1) + 1], F[2 * i])

		tol_ss = 1e-10
		ttss = - np.log(tol_ss) / zeta / omega_n
		N_ss = int(ttss / dt)
		X = X[N_ss:]
		V = V[N_ss:]
		F = np.insert(F,0,0)
		F = F[N_ss:][::2]

		xmax = .25
		vmax = 0.6
		pX_i, x_edge = np.histogram(X, bins=pX.size, range=(-xmax, xmax), normed=True)
		x = x_edge[:-1] + (x_edge[1] - x_edge[0]) / 2
		pV_i, v_edge = np.histogram(V, bins=pV.size, range=(-vmax, vmax), normed=True)
		v = v_edge[:-1] + (v_edge[1] - v_edge[0]) / 2
		pF_i, f_edge = np.histogram(F[::2], bins=pF.size, range=(-3 * sigma_F, 3 * sigma_F), normed=True)
		f = f_edge[:-1] + (f_edge[1] - f_edge[0]) / 2
		H_i, xedges, vedges = np.histogram2d(X, V, range=[[-xmax, xmax], [-vmax, vmax]], bins=H.shape, normed=True)
		pX += pX_i / float(loops)
		pV += pV_i / float(loops)
		pF += pF_i / float(loops)
		H += H_i / float(loops)
		sigma_x_meas += np.std(X) / loops
		sigma_v_meas += np.std(V) / loops
	print 'STD X: ', sigma_x_meas
	print 'STD V: ', sigma_v_meas
	dx = x[1] - x[0]
	dv = v[1] - v[0]
	var_meas = (np.sum(x**2 * pX) * dx)
	var, e = quad(response_displ_AR1, 0, np.pi / dt, args=(dt, AR1, sigma_w))
	print np.std(X)
	print '------------------------'
	print 'Exact STD:', var**.5
	print 'Measured STD: ', var_meas**.5
	print 'error: ', var_meas**.5 / var**.5 * 100 - 100
	print '------------------------'
	# Time samples
	np.savez('displacement', x=x, array=pX)
	np.savez('velocity', v=v, array=pV)
	np.savez('joint', x=x, v=v, array=H)
	np.savez('force', f=f, array=pF)
	M = 10000
	M = np.min([M, len(X)])
	t = (np.arange(M) + N_it / 2) * dt
	fig1, (ax1, ax2, ax3) = plt.subplots(3, sharex=True, figsize=(10,7))

	ax1.hlines(0, t[0], t[-1], color='silver', lw=1.5)
	ax2.hlines(0, t[0], t[-1], color='silver', lw=1.5)
	ax2.hlines(0, t[0], t[-1], color='silver', lw=1.5)
	ax1.plot(t, X[:M])
	ax2.plot(t, V[: M])
	ax3.plot(t, F[: M])

	ax1.set_title('Displacement')
	ax2.set_title('Velocity')
	ax3.set_title('Force')
	ax3.set_xlabel('Seconds')

	# joint distr
	Z = H.T
	vmax = 13
	ndig = 2
	fig2, ax = plt.subplots(figsize=(10,7))
	cax = ax.imshow(Z, interpolation='nearest', extent=[xedges[0], xedges[-1], vedges[0], vedges[-1]], aspect='auto', vmin=0, vmax=vmax)
	c_max = round(vmax, ndig)
	c_mid = c_max / 2.
	cbar = fig2.colorbar(cax, ticks=[0, c_mid, vmax])
	cbar.ax.set_yticklabels(['0.0', c_mid, c_max])  # vertically oriented colorbar
	ax.set_xlabel('Displacement')
	ax.set_ylabel('Velocity')
	# Distribution plots B
	fig3, (ax1, ax2, ax3) = plt.subplots(3, figsize=(10,7))
	fig3.tight_layout()
	x = x_edge[:-1] + (x_edge[1] - x_edge[0]) / 2
	v = v_edge[:-1] + (v_edge[1] - v_edge[0]) / 2
	f = f_edge[:-1] + (f_edge[1] - f_edge[0]) / 2
	# exact distributions
	print "STD_F = ", np.std(F)

	exact_dist = False
	if exact_dist:
		xx = np.linspace(- 3.5 * sigma_x, 3.5 * sigma_x, 1000)
		vv = np.linspace(- 3.5 * sigma_v, 3.5 * sigma_v, 1000)
		ff = np.linspace(- 3.5 * sigma_F, 3.5 * sigma_F, 1000)
		pXX = ss.norm.pdf(xx, scale=sigma_x)
		pVV = ss.norm.pdf(vv, scale=sigma_v)
		pFF = ss.norm.pdf(ff, scale=sigma_F)
		ax1.plot(xx, pXX, lw=2, alpha=.8, c='k', label='std= %s' % (round(sigma_x, 6)))
		ax2.plot(vv, pVV, lw=2, alpha=.8, c='k', label='std= %s' % (round(sigma_v, 6)))
		ax3.plot(ff, pFF, lw=2, alpha=.8, c='k', label='std= %s' % (int(sigma_F)))
	ax1.plot(x, pX)
	ax2.plot(v, pV)
	ax3.plot(f, pF)
	ax1.set_title('Displacement')
	ax2.set_title('Velocity')
	ax3.set_title('Force')
	ax1.legend(fancybox=True)
	ax2.legend(fancybox=True)
	ax3.legend(fancybox=True)
	########################################################################
	mesh, par, p = load_array('whitsine1')
	print par

	xx = mesh['x']
	vv = mesh['v']
	dx = x[1] - x[0]
	dv = v[1] - v[0]
	PX, PV, PXV = plot.data(p, mesh)
	fig0, ax = plt.subplots(figsize=(10, 7), dpi=300)
	ax.plot(xx, PX / dx, label='PDE', lw=1.5)
	ax.plot(x, pX, label='Monte Carlo', lw = 1.5, c='r')
	ax.legend(fancybox=True)
	figg0, ax = plt.subplots(figsize=(10, 7))
	ax.plot(vv, PV / dv, label='PDE', lw=1.5)
	ax.plot(v, pV, label='Monte Carlo', lw=1.5, c='r')
	ax.legend(fancybox=True)



	#######################################################################
	# response spectrum
	fig, (ax1, ax2, ax3) = plt.subplots(3, sharex=True, figsize=(10,7))
	w = np.linspace(0, np.pi / dt, 1000)
	if 'AR' in par.keys():
		if len(par['AR']) == 1:
			ax1.plot(w, response_displ_AR1(w, dt, AR1, sigma_w), c='k', lw=2, alpha=.8, label='exact')
			ax2.plot(w, response_vel_AR1(w, dt, AR1, sigma_w), c='k', lw=2, alpha=.8, label='exact')
			ax3.plot(w, sigma_w**2 * dt / np.pi / (1 + AR1**2 - 2 * AR1 * np.cos(w * dt)), c='k', lw=2, alpha=.8, label='exact')
		
		elif len(par['AR']) == 2:
			ax1.plot(w, response_displ_AR2(w, dt, AR1, AR2, sigma_w), c='k', lw=2, alpha=.8, label='exact')
			ax2.plot(w, response_vel_AR2(w, dt, AR1, AR2, sigma_w), c='k', lw=2, alpha=.8, label='exact')
			ax3.plot(w, sigma_w**2 * dt / np.pi / (1. + AR1**2 + AR2**2 - 2. * AR1 * (1. - AR2) * np.cos(w * dt) - 2 * AR2 * np.cos(2 * w * dt)), c='k', lw=2, alpha=.8, label='exact')

	f, Sxx = welch(X, 1. / dt, nperseg=5000)
	f, Svv = welch(V, 1. / dt, nperseg=5000)
	ax1.semilogy(f * np.pi * 2, Sxx / 2. / np.pi, label='DFT')
	ax2.semilogy(f * np.pi * 2, Svv / 2 / np.pi)
	f, SFF = welch(F, 1. / dt, nperseg=2000)
	ax3.semilogy(f * 2 * np.pi, SFF / 2 / np.pi)
	ax1.set_title('Displacement spectrum')
	ax2.set_title('Velocity spectrum')
	ax3.set_title('Force spectrum')
	ax3.set_xlabel('$\omega$', size=15)
	ax1.legend(fancybox=True)
	plt.show()
	print max(Sxx)


N = int(1e6)
simulate(N, noise=1)

