from main import load_sparse_csc, save_array, load_array
import plotting as plot
import matplotlib.pyplot as plt
import numpy as np
from scipy.sparse import linalg
import time
from scipy.linalg import norm


def iterate_steadystate(Matrix, p0=0, max_it=0, tol=0):
	'''
	Matrix : Probability matrix
	max_it : maximum number of iterations. Default 10x vector lenght
	tol : how big change in next state is tolerated. default = nummerical precision.
	'''
	n, n = P.shape
	print "size: ", n
	if max_it == 0:
		max_it = n * 10
	if tol == 0:
		tol = 1e-5
	t0 = time.time()
	if type(p0) == int:
		pi1 = np.random.rand(n)
		pi = pi1 / np.sum(pi1)
	else:
		pi = p0
		pi1 = np.ones(n)
	e = np.zeros(max_it / 2)
	n = norm(pi - pi1, 1); i = 0
	print "No. Iterations:"
	while n > tol and i < max_it / 2:
		pi1 = P.dot(pi)
		pi = P.dot(pi1)
		n = norm(pi - pi1, 1)
		e[i] = n
		i += 1
		if i%50 == 0:
			print i*2
	t = time.time() - t0
	fig, ax = plt.subplots()
	ax.semilogy(np.arange(e.size) * 2, e)
	ax.set_xlabel('Iterations')
	ax.set_ylabel('error')
	print 'Number of iterations done: ', i*2
	print '%s Iterations took %s seconds' % (i, t)
	return pi


def eigenval_steadystate(P, p0=0, max_it=0, tol=0):
	'''
	Matrix : Probability matrix
	p0 : starting vector, if p0=0 a random vector i chosen.
	max_it : maximum number of iterations. Default 10x vector lenght
	tol : tolerance in range 0 -> 1.
	'''
	n, n = P.shape
	if type(p0) == int:
		p = np.random.rand(n)
		p = p / p.sum()
	else:
		p = p0
	if max_it == 0:
		max_it = n * 10
	t0 = time.time()
	lamda, p = linalg.eiglinals(P, k=1, v0=p, which='LM', tol=tol, maxiter=max_it)
	t = time.time() - t0
	print 'lambda = ', lamda
	print 'Eigensolving arnoldi took %s seconds' % (t)
	if np.isreal(p).all():
		p = np.real(p)
		return p / p.sum()
	else:
		print 'p contains complex values'
		return p


filename = 'WN'


mesh, par, P = load_sparse_csc(filename)

try:
	mesh, par, p = load_array(filename)
except IOError:
	p = iterate_steadystate(P, max_it=4000, tol=1e-5)
	# p = eigenval_steadystate(P, p0=p, tol=1e-5)
	save_array(filename, p, mesh, par)


smoothen = 0
plot.colormap(p, mesh, smoothen=smoothen, ndig=2, vmax=0)
plot.distribution(p, mesh, smoothen=smoothen)
plot.plot3D(p, mesh, smoothen=smoothen)
plt.show()
